import re

from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator, RegexValidator
from django.db import models
from django.forms.models import ModelForm, ModelFormMetaclass
from django.utils.translation import gettext_lazy as _

from servala.core.models import ServiceInstance


class CRDModel(models.Model):
    """Base class for all virtual CRD models"""

    def __init__(self, **kwargs):
        if spec := kwargs.pop("spec", None):
            kwargs.update(unnest_data({"spec": spec}))
        super().__init__(**kwargs)

    class Meta:
        abstract = True


def duplicate_field(field_name, model):
    # Get the field from the model
    field = model._meta.get_field(field_name)

    # Create a new field with the same attributes
    new_field = type(field).__new__(type(field))
    new_field.__dict__.update(field.__dict__)

    # Ensure the field is not linked to the original model
    new_field.model = None
    new_field.auto_created = False

    return new_field


def generate_django_model(schema, group, version, kind):
    """
    Generates a virtual Django model from a Kubernetes CRD's OpenAPI v3 schema.
    """
    # We always need these three fields to know our own name and our full namespace
    model_fields = {"__module__": "crd_models"}
    for field_name in ("name", "organization", "context"):
        model_fields[field_name] = duplicate_field(field_name, ServiceInstance)

    # All other fields are generated from the schema, except for the
    # resourceRef object
    spec = schema["properties"].get("spec") or {}
    spec["properties"].pop("resourceRef", None)
    model_fields.update(build_object_fields(spec, "spec", parent_required=True))

    # Store the original schema on the model class
    model_fields["SCHEMA"] = schema

    meta_class = type("Meta", (), {"app_label": "crd_models"})
    model_fields["Meta"] = meta_class

    # create the model class
    model_name = kind
    model_class = type(model_name, (CRDModel,), model_fields)
    return model_class


def build_object_fields(schema, name, verbose_name_prefix=None, parent_required=False):
    required_fields = schema.get("required") or []
    properties = schema.get("properties") or {}
    fields = {}

    for field_name, field_schema in properties.items():
        is_required = field_name in required_fields and parent_required
        full_name = f"{name}.{field_name}"
        result = get_django_field(
            field_schema,
            is_required,
            field_name,
            full_name,
            verbose_name_prefix=verbose_name_prefix,
        )
        if isinstance(result, dict):
            fields.update(result)
        else:
            fields[full_name] = result
    return fields


def deslugify(title):
    if "_" in title:
        title.replace("_", " ")
        return title.title()
    return re.sub(r"(?<!^)(?=[A-Z])", " ", title).capitalize()


def get_django_field(
    field_schema, is_required, field_name, full_name, verbose_name_prefix=None
):
    field_type = field_schema.get("type") or "string"
    format = field_schema.get("format")
    verbose_name_prefix = verbose_name_prefix or ""
    verbose_name = f"{verbose_name_prefix} {deslugify(field_name)}".strip()

    # Pass down the requirement status from parent to child fields
    kwargs = {
        "blank": not is_required,  # All fields are optional by default
        "null": not is_required,
        "help_text": field_schema.get("description"),
        "validators": [],
        "verbose_name": verbose_name,
        "default": field_schema.get("default"),
    }

    if minimum := field_schema.get("minimum"):
        kwargs["validators"].append(MinValueValidator(minimum))
    if maximum := field_schema.get("maximum"):
        kwargs["validators"].append(MaxValueValidator(maximum))

    if field_type == "string":
        if format == "date-time":
            return models.DateTimeField(**kwargs)
        elif format == "date":
            return models.DateField(**kwargs)
        else:
            max_length = field_schema.get("max_length") or 255
            if pattern := field_schema.get("pattern"):
                kwargs["validators"].append(RegexValidator(regex=pattern))
            if choices := field_schema.get("enum"):
                kwargs["choices"] = ((choice, choice) for choice in choices)
            return models.CharField(max_length=max_length, **kwargs)
    elif field_type == "integer":
        return models.IntegerField(**kwargs)
    elif field_type == "number":
        return models.FloatField(**kwargs)
    elif field_type == "boolean":
        return models.BooleanField(**kwargs)
    elif field_type == "object":
        # Here we pass down the requirement status to nested objects
        return build_object_fields(
            field_schema,
            full_name,
            verbose_name_prefix=f"{verbose_name}:",
            parent_required=is_required,
        )
    elif field_type == "array":
        # TODO: handle items / validate items, build multi-select input
        # if field_schema.get("items") and (choices := field_schema["items"].get("enum")):
        #   choices = [c, c for c in choices]
        kwargs["help_text"] = _("JSON field (array)")
        return models.JSONField(**kwargs)
    return models.CharField(max_length=255, **kwargs)


def unnest_data(data):
    result = {}

    def _flatten_dict(d, parent_key=""):
        for key, value in d.items():
            new_key = f"{parent_key}.{key}" if parent_key else key
            if isinstance(value, dict):
                _flatten_dict(value, new_key)
            else:
                result[new_key] = value

    _flatten_dict(data)
    return result


class CrdModelFormMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.schema = self._meta.model.SCHEMA

        for field in ("organization", "context"):
            self.fields[field].widget = forms.HiddenInput()

        if self.instance and self.instance.pk:
            self.fields["name"].disabled = True
            self.fields["name"].help_text = _("Name cannot be changed after creation.")
            self.fields["name"].widget = forms.HiddenInput()

    def strip_title(self, field_name, label):
        field = self.fields[field_name]
        if field and field.label.startswith(label):
            field.label = field.label[len(label) :]

    def get_fieldsets(self):
        fieldsets = []

        # General fieldset for non-spec fields
        general_fields = [
            field_name
            for field_name, field in self.fields.items()
            if not field_name.startswith("spec.")
        ]
        if general_fields:
            fieldset = {"title": "General", "fields": general_fields, "fieldsets": []}
            if all(
                [
                    isinstance(self.fields[field].widget, forms.HiddenInput)
                    for field in general_fields
                ]
            ):
                fieldset["hidden"] = True
            fieldsets.append(fieldset)

        # Process spec fields
        others = []
        nested_fieldsets = {}

        for field_name in self.fields:
            if field_name.startswith("spec."):
                parts = field_name.split(".")
                if len(parts) == 2:  # Top-level spec field
                    others.append(field_name)
                else:
                    parent_key = parts[1]
                    if not nested_fieldsets.get(parent_key):
                        nested_fieldsets[parent_key] = {
                            "fields": [],
                            "fieldsets": {},
                            "title": deslugify(parent_key),
                        }
                    parent = nested_fieldsets[parent_key]
                    if len(parts) == 3:  # Top-level within fieldset
                        parent["fields"].append(field_name)
                    else:
                        sub_key = parts[2]
                        if not parent["fieldsets"].get(sub_key):
                            parent["fieldsets"][sub_key] = {
                                "title": deslugify(sub_key),
                                "fields": [],
                            }
                        parent["fieldsets"][sub_key]["fields"].append(field_name)

        # Add nested fieldsets to fieldsets
        for group in nested_fieldsets.values():
            total_fields = 0
            for fieldset in group["fieldsets"].values():
                if (field_count := len(fieldset["fields"])) == 1:
                    group["fields"].append(fieldset["fields"][0])
                    fieldset["fields"] = []
                else:
                    title = f"{group['title']}: {fieldset['title']}: "
                    for field in fieldset["fields"]:
                        self.strip_title(field, title)
                    total_fields += field_count
            if (total_fields + len(group["fields"])) == 1:
                others.append(group["fields"][0])
            else:
                title = f"{group['title']}: "
                for field in group["fields"]:
                    self.strip_title(field, title)
                fieldsets.append(group)

        # Add 'others' tab if there are any fields
        if others:
            fieldsets.append({"title": "Others", "fields": others, "fieldsets": []})

        return fieldsets

    def get_nested_data(self):
        """
        Builds the original nested JSON structure from flat form data.
        Form fields are named with dot notation (e.g., 'spec.replicas')
        """
        result = {}

        for field_name, value in self.cleaned_data.items():
            if value is None or value == "":
                continue

            parts = field_name.split(".")
            current = result

            # Navigate through the nested structure
            for i, part in enumerate(parts):
                if i == len(parts) - 1:
                    # Last part, set the value
                    current[part] = value
                else:
                    # Create nested dict if it doesn't exist
                    if part not in current:
                        current[part] = {}
                    current = current[part]
        return result

    def clean(self):
        cleaned_data = super().clean()
        self.validate_nested_data(
            self.get_nested_data().get("spec", {}), self.schema["properties"]["spec"]
        )
        return cleaned_data

    def validate_nested_data(self, data, schema):
        """Validate data against the provided OpenAPI v3 schema"""
        # TODO: actually validate the nested data.
        # TODO: get jsonschema to give us a path to the failing field rather than just an error message,
        # then add the validation error to that field (self.add_error())
        # try:
        #     validate(instance=data, schema=schema)
        # except Exception as e:
        #     raise forms.ValidationError(f"Validation error: {e.message}")
        pass


def generate_model_form_class(model):
    meta_attrs = {
        "model": model,
        "fields": "__all__",
    }
    fields = {
        "Meta": type("Meta", (object,), meta_attrs),
        "__module__": "crd_models",
    }
    class_name = f"{model.__name__}ModelForm"
    return ModelFormMetaclass(class_name, (CrdModelFormMixin, ModelForm), fields)
